@@ -16,7 +16,7 @@
  *   You should have received a copy of the GNU General Public License     *
  *   along with this program; if not, write to the                         *
  *   Free Software Foundation, Inc.,                                       *
- *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
+ *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.             *
  ***************************************************************************/
 #include <qdir.h>
 #include <qfile.h>
@@ -30,8 +30,13 @@
 #include <util/functions.h>
 #include <util/fileops.h>
 #include <interfaces/functions.h>
+#include <interfaces/trackerslist.h>
+#include <datachecker/singledatachecker.h>
+#include <datachecker/multidatachecker.h>
+#include <datachecker/datacheckerlistener.h>
 #include <migrate/ccmigrate.h>
 #include <migrate/cachemigrate.h>
+#include <kademlia/dhtbase.h>
 #include "downloader.h"
 #include "uploader.h"
 #include "tracker.h"
@@ -53,6 +58,9 @@
 #include "downloadcap.h"
 #include "uploadcap.h"
 #include "queuemanager.h"
+#include "statsfile.h"
+#include "announcelist.h"
+#include "preallocationthread.h"
 
 using namespace kt;
 
@@ -60,26 +68,34 @@ namespace bt
 {
 
 
+
 	TorrentControl::TorrentControl()
-			: tor(0),tracker(0),cman(0),pman(0),down(0),up(0),choke(0),tmon(0)
+	: tor(0),tracker(0),cman(0),pman(0),down(0),up(0),choke(0),tmon(0),prealloc(false), last_announce(0)
 	{
 		stats.imported_bytes = 0;
+		stats.trk_bytes_downloaded = 0;
+		stats.trk_bytes_uploaded = 0;
 		stats.running = false;
 		stats.started = false;
 		stats.stopped_by_error = false;
 		stats.session_bytes_downloaded = 0;
 		stats.session_bytes_uploaded = 0;
+		session_bytes_uploaded = 0;
 		old_datadir = QString::null;
 		stats.status = NOT_STARTED;
 		stats.autostart = true;
 		stats.user_controlled = false;
+		stats.priv_torrent = false;
 		running_time_dl = running_time_ul = 0;
 		prev_bytes_dl = 0;
 		prev_bytes_ul = 0;
+		trk_prev_bytes_dl = trk_prev_bytes_ul = 0;
 		io_error = false;
 		priority = 0;
-
+		maxShareRatio = 0.00f;
+		custom_output_name = false;
 		updateStats();
+		prealoc_thread = 0;
 	}
 
 
@@ -89,7 +105,7 @@ namespace bt
 	{
 		if (stats.running)
 			stop(false);
-
+		
 		if (tmon)
 			tmon->destroyed();
 		delete choke;
@@ -103,8 +119,8 @@ namespace bt
 
 	void TorrentControl::update()
 	{
-		// do not update during critical operation mode
-		if (Globals::instance().inCriticalOperationMode())
+		// do not update during critical operation mode or when we are data checking
+		if (Globals::instance().inCriticalOperationMode() || stats.status == kt::CHECKING_DATA)
 			return;
 		
 		if (io_error)
@@ -114,6 +130,35 @@ namespace bt
 			return;
 		}
 		
+		if (prealoc_thread)
+		{
+			if (prealoc_thread->isDone())
+			{
+				// thread done
+				if (prealoc_thread->errorHappened())
+				{
+					// upon error just call onIOError and return
+					onIOError(prealoc_thread->errorMessage());
+					delete prealoc_thread;
+					prealoc_thread = 0;
+					prealloc = true; // still need to do preallocation
+					return;
+				}
+				else
+				{
+					// continue the startup of the torrent
+					delete prealoc_thread;
+					prealoc_thread = 0;
+					prealloc = false;
+					stats.status = kt::NOT_STARTED;
+					saveStats();
+					continueStart();
+				}
+			}
+			else
+				return; // preallocation still going on, so just return
+		}
+		
 		try
 		{
 			// first update peermanager
@@ -128,12 +173,17 @@ namespace bt
 			stats.completed = cman->chunksLeft() == 0;
 			if (stats.completed && !comp)
 			{
-				// download has just been completed
-				tracker->completed();
 				pman->killSeeders();
 				QDateTime now = QDateTime::currentDateTime();
 				running_time_dl += time_started_dl.secsTo(now);
 				updateStatusMsg();
+				updateStats();
+				
+				// download has just been completed
+				// only sent completed to tracker when we have all chunks (so no excluded chunks)
+				if (cman->haveAllChunks())
+					tracker->completed();
+				
 				finished(this);
 			}
 			else if (!stats.completed && comp)
@@ -172,19 +222,25 @@ namespace bt
 			}
 
 			// Update DownloadCap
-			DownloadCap::instance().update(stats.download_rate);
-			UploadCap::instance().update();
 			updateStats();
 			if (stats.download_rate > 0)
 				stalled_timer.update();
 			
 			// do a manual update if we are stalled for more then 2 minutes
-			if (stalled_timer.getElapsedSinceUpdate() > 120000 && stats.bytes_left > 0)
+			// we do not do this for private torrents
+			if (stalled_timer.getElapsedSinceUpdate() > 120000 && !stats.completed && !stats.priv_torrent)
 			{
 				Out() << "Stalled for to long, time to get some fresh blood" << endl;
 				tracker->manualUpdate();
 				stalled_timer.update();
 			}
+			
+			if(overMaxRatio())
+			{
+				stop(true);
+				emit seedingAutoStopped(this);
+			}
+			
 		}
 		catch (Error & e)
 		{
@@ -203,13 +259,27 @@ namespace bt
 	}
 
 	void TorrentControl::start()
-	{
-		if (bt::Exists(datadir + "stopped"))
-			bt::Delete(datadir + "stopped",true);
+	{	
+		// do not start running torrents
+		if (stats.running || stats.status == kt::ALLOCATING_DISKSPACE)
+			return;
 
 		stats.stopped_by_error = false;
 		io_error = false;
-		pman->start();
+		try
+		{
+			bool ret = true;
+			aboutToBeStarted(this,ret);
+			if (!ret)
+				return;
+		}
+		catch (Error & err)
+		{
+			// something went wrong when files were recreated, set error and rethrow
+			onIOError(err.toString());
+			return;
+		}
+		
 		try
 		{
 			cman->start();
@@ -220,6 +290,29 @@ namespace bt
 			throw;
 		}
 		
+		time_started_ul = time_started_dl = QDateTime::currentDateTime();
+		trk_prev_bytes_dl = stats.bytes_downloaded,
+		trk_prev_bytes_ul = stats.bytes_uploaded,
+		stats.trk_bytes_downloaded = 0;
+		stats.trk_bytes_uploaded = 0;
+		
+		if (prealloc)
+		{
+			Out(SYS_GEN|LOG_NOTICE) << "Pre-allocating diskspace" << endl;
+			prealoc_thread = new PreallocationThread(cman);
+			stats.running = true;
+			stats.status = kt::ALLOCATING_DISKSPACE;
+			prealoc_thread->start();
+			return;
+		}
+		
+		continueStart();
+	}
+	
+	void TorrentControl::continueStart()
+	{
+		// continues start after the prealoc_thread has finished preallocation
+		pman->start();
 		try
 		{
 			down->loadDownloads(datadir + "current_chunks");
@@ -228,7 +321,7 @@ namespace bt
 		{
 			// print out warning in case of failure
 			// we can still continue the download
-			Out() << "Warning : " << e.toString() << endl;
+			Out(SYS_GEN|LOG_NOTICE) << "Warning : " << e.toString() << endl;
 		}
 		
 		loadStats();
@@ -237,10 +330,13 @@ namespace bt
 		stats.autostart = true;
 		choker_update_timer.update();
 		stats_save_timer.update();
+		
+		
+		stalled_timer.update();
 		tracker->start();
-		time_started_ul = time_started_dl = QDateTime::currentDateTime();
 		stalled_timer.update();
 	}
+		
 
 	void TorrentControl::stop(bool user)
 	{
@@ -249,6 +345,27 @@ namespace bt
 			running_time_dl += time_started_dl.secsTo(now);
 		running_time_ul += time_started_ul.secsTo(now);
 		time_started_ul = time_started_dl = now;
+		
+		// stop preallocation thread if necesarry
+		if (prealoc_thread)
+		{
+			prealoc_thread->stop();
+			prealoc_thread->wait();
+			
+			if (prealoc_thread->errorHappened() || prealoc_thread->isNotFinished())
+			{
+				delete prealoc_thread;
+				prealoc_thread = 0;
+				prealloc = true;
+				saveStats(); // save stats, so that we will start preallocating the next time
+			}
+			else
+			{
+				delete prealoc_thread;
+				prealoc_thread = 0;
+				prealloc = false;
+			}
+		}
 	
 		if (stats.running)
 		{
@@ -285,6 +402,8 @@ namespace bt
 		saveStats();
 		updateStatusMsg();
 		updateStats();
+		stats.trk_bytes_downloaded = 0;
+		stats.trk_bytes_uploaded = 0;
 	}
 
 	void TorrentControl::setMonitor(kt::MonitorInterface* tmo)
@@ -298,7 +417,11 @@ namespace bt
 		}
 	}
 
-	void TorrentControl::init(const QueueManager* qman,const QString & torrent,const QString & tmpdir,const QString & ddir,const QString & default_save_dir)
+	void TorrentControl::init(QueueManager* qman,
+							  const QString & torrent,
+							  const QString & tmpdir,
+							  const QString & ddir,
+							  const QString & default_save_dir)
 	{
 		datadir = tmpdir;
 		stats.completed = false;
@@ -327,7 +450,17 @@ namespace bt
 		// check if we haven't already loaded the torrent
 		// only do this when qman isn't 0
 		if (qman && qman->allreadyLoaded(tor->getInfoHash()))
-			throw Error(i18n("You are already downloading this torrent."));
+		{
+			if (tor->getAnnounceList())
+			{
+				qman->mergeAnnounceList(tor->getInfoHash(),tor->getAnnounceList());
+			}
+			else 
+			{
+				qman->addTrackerURL(tor->getInfoHash(),tor->getTrackerURL(true));
+			}
+			throw Error(i18n("You are already downloading this torrent, the list of trackers of both torrents has been merged."));
+		}
 		
 		if (!bt::Exists(datadir))
 		{
@@ -337,11 +470,19 @@ namespace bt
 		stats.torrent_name = tor->getNameSuggestion();
 		stats.multi_file_torrent = tor->isMultiFile();
 		stats.total_bytes = tor->getFileLength();
+		stats.priv_torrent = tor->isPrivate();
 		
-		// load stats if outputdir is null
+		// check the stats file for the custom_output_name variable
+		StatsFile st(datadir + "stats");
+		if (st.hasKey("CUSTOM_OUTPUT_NAME") && st.readULong("CUSTOM_OUTPUT_NAME") == 1)
+		{
+			custom_output_name = true;
+		}
+		
+		// load outputdir if outputdir is null
 		if (outputdir.isNull() || outputdir.length() == 0)
 			loadOutputDir();
-		
+					
 		// copy torrent in temp dir
 		QString tor_copy = datadir + "torrent";
 
@@ -369,29 +510,30 @@ namespace bt
 
 		// create PeerManager and Tracker
 		pman = new PeerManager(*tor);
-		if (tor->getTrackerURL(true).protocol() == "udp")
-			tracker = new UDPTracker(this,tor->getInfoHash(),tor->getPeerID());
-		else
-			tracker = new HTTPTracker(this,tor->getInfoHash(),tor->getPeerID());
-
+		KURL url = tor->getTrackerURL(true);
+		//Out() << "Tracker url " << url << " " << url.protocol() << " " << url.prettyURL() << endl;
+		tracker = new Tracker(this,tor->getInfoHash(),tor->getPeerID());
 		connect(tracker,SIGNAL(error()),this,SLOT(trackerResponseError()));
 		connect(tracker,SIGNAL(dataReady()),this,SLOT(trackerResponse()));
 
 
 		// Create chunkmanager, load the index file if it exists
 		// else create all the necesarry files
-		cman = new ChunkManager(*tor,datadir,outputdir);
+		cman = new ChunkManager(*tor,datadir,outputdir,custom_output_name);
 		// outputdir is null, see if the cache has figured out what it is
 		if (outputdir.length() == 0)
 			outputdir = cman->getDataDir();
 		
+		// store the outputdir into the output_path variable, so others can access it	
+		
 		connect(cman,SIGNAL(updateStats()),this,SLOT(updateStats()));
 		if (bt::Exists(datadir + "index"))
 			cman->loadIndexFile();
 
-		//if (!bt::Exists(datadir + "cache"))
-		// as a sanity check make sure all files are created properly
-		cman->createFiles();
+	
+		// as a sanity check make sure all files are created properly the first time we load this torrent
+		if (!stats.completed && !ddir.isNull())
+			cman->createFiles();
 
 		stats.completed = cman->chunksLeft() == 0;
 
@@ -400,13 +542,13 @@ namespace bt
 		connect(down,SIGNAL(ioError(const QString& )),
 				this,SLOT(onIOError(const QString& )));
 		up = new Uploader(*cman,*pman);
-		choke = new Choker(*pman);
+		choke = new Choker(*pman,*cman);
 
 
 		connect(pman,SIGNAL(newPeer(Peer* )),this,SLOT(onNewPeer(Peer* )));
 		connect(pman,SIGNAL(peerKilled(Peer* )),this,SLOT(onPeerRemoved(Peer* )));
-		connect(cman,SIGNAL(excluded(Uint32, Uint32 )),
-		        down,SLOT(onExcluded(Uint32, Uint32 )));
+		connect(cman,SIGNAL(excluded(Uint32, Uint32 )),down,SLOT(onExcluded(Uint32, Uint32 )));
+		connect(cman,SIGNAL(included( Uint32, Uint32 )),down,SLOT(onIncluded( Uint32, Uint32 )));
 
 		updateStatusMsg();
 
@@ -431,6 +573,14 @@ namespace bt
 		loadStats();
 		updateStats();
 		saveStats();
+		stats.output_path = cman->getOutputPath();
+		Out() << "OutputPath = " << stats.output_path << endl;
+		
+		AnnounceList* list = tor->getAnnounceList();
+		if(!list)
+			list = tor->createAnnounceList();
+		
+		list->setDatadir(datadir);
 	}
 
 	void TorrentControl::trackerResponse()
@@ -443,7 +593,7 @@ namespace bt
 		}
 		catch (Error & e)
 		{
-			Out() << "Error : " << e.toString() << endl;
+			Out(SYS_TRK|LOG_IMPORTANT) << "Error : " << e.toString() << endl;
 			stats.trackerstatus = i18n("Invalid response");
 			tracker->handleError();
 		}
@@ -451,45 +601,84 @@ namespace bt
 
 	void TorrentControl::trackerResponseError()
 	{
-		Out() << "Tracker Response Error" << endl;
+		Out(SYS_TRK|LOG_IMPORTANT) << "Tracker Response Error" << endl;
 		stats.trackerstatus = i18n("Unreachable");
 		tracker->handleError();
 	}
 
+	bool TorrentControl::announceAllowed()
+	{
+		if(last_announce == 0)
+			return true;
+		
+		return bt::GetCurrentTime() - last_announce >= 60 * 1000;
+	}
+	
 	void TorrentControl::updateTracker()
 	{
-		if (stats.running)
+		if (stats.running && announceAllowed())
+		{
 			tracker->manualUpdate();
+			last_announce = bt::GetCurrentTime();
+		}
 	}
 
 	KURL TorrentControl::getTrackerURL(bool prev_success) const
 	{
-		return tor->getTrackerURL(prev_success);
+		if (tor)
+			return tor->getTrackerURL(prev_success);
+		else
+			return KURL();
 	}
 
-
 	void TorrentControl::onNewPeer(Peer* p)
 	{
-		p->getPacketWriter().sendBitSet(cman->getBitSet());
+		connect(p,SIGNAL(gotPortPacket( const QString&, Uint16 )),
+				this,SLOT(onPortPacket( const QString&, Uint16 )));
+		
+		if (p->getStats().fast_extensions)
+		{
+			const BitSet & bs = cman->getBitSet();
+			if (bs.allOn())
+				p->getPacketWriter().sendHaveAll();
+			else if (bs.numOnBits() == 0)
+				p->getPacketWriter().sendHaveNone();
+			else
+				p->getPacketWriter().sendBitSet(bs);
+		}
+		else
+		{
+			p->getPacketWriter().sendBitSet(cman->getBitSet());
+		}
+		
 		if (!stats.completed)
 			p->getPacketWriter().sendInterested();
+		
+		if (p->isDHTSupported() && !stats.priv_torrent)
+			p->getPacketWriter().sendPort(Globals::instance().getDHT().getPort());
+		
 		if (tmon)
 			tmon->peerAdded(p);
 	}
 
 	void TorrentControl::onPeerRemoved(Peer* p)
 	{
+		disconnect(p,SIGNAL(gotPortPacket( const QString&, Uint16 )),
+				this,SLOT(onPortPacket( const QString&, Uint16 )));
 		if (tmon)
 			tmon->peerRemoved(p);
 	}
 
 	void TorrentControl::doChoking()
 	{
-		choke->update(cman->bytesLeft() == 0);
+		choke->update(stats.completed,stats);
 	}
 
 	bool TorrentControl::changeDataDir(const QString & new_dir)
 	{
+#warning "This code is broken at the moment, will fix later"
+		return false;
+
 		// new_dir doesn't contain the torX/ part
 		// so first get that and append it to new_dir
 		int dd = datadir.findRev(DirSeparator(),datadir.length() - 2,false);
@@ -572,6 +761,8 @@ namespace bt
 			stats.status = kt::ERROR;
 		else if (!stats.started)
 			stats.status = kt::NOT_STARTED;
+		else if(!stats.running && !stats.user_controlled)
+			stats.status = kt::QUEUED;
 		else if (!stats.running && stats.completed)
 			stats.status = kt::COMPLETE;
 		else if (!stats.running)
@@ -609,138 +800,79 @@ namespace bt
 
 	void TorrentControl::saveStats()
 	{
-		QFile fptr(datadir + "stats");
-		if (!fptr.open(IO_WriteOnly))
-		{
-			Out() << "Warning : can't create stats file" << endl;
-			return;
-		}
+		StatsFile st(datadir + "stats");
 
-		QTextStream out(&fptr);
-		out << "OUTPUTDIR=" << cman->getDataDir() << ::endl;
+		st.write("OUTPUTDIR", cman->getDataDir());
+		
 		if (cman->getDataDir() != outputdir)
 			outputdir = cman->getDataDir();
-		out << "UPLOADED=" << QString::number(up->bytesUploaded()) << ::endl;
+		
+		st.write("UPLOADED", QString::number(up->bytesUploaded()));
+		
 		if (stats.running)
 		{
 			QDateTime now = QDateTime::currentDateTime();
-			out << "RUNNING_TIME_DL=" << (running_time_dl + time_started_dl.secsTo(now)) << ::endl;
-			out << "RUNNING_TIME_UL=" << (running_time_ul + time_started_ul.secsTo(now)) << ::endl;
+			st.write("RUNNING_TIME_DL",QString("%1").arg(running_time_dl + time_started_dl.secsTo(now)));
+			st.write("RUNNING_TIME_UL",QString("%1").arg(running_time_ul + time_started_ul.secsTo(now)));
 		}
 		else
 		{
-			out << "RUNNING_TIME_DL=" << running_time_dl << ::endl;
-			out << "RUNNING_TIME_UL=" << running_time_ul << ::endl;
+			st.write("RUNNING_TIME_DL", QString("%1").arg(running_time_dl));
+			st.write("RUNNING_TIME_UL", QString("%1").arg(running_time_ul));
 		}
 		
-		out << "PRIORITY=" << priority << ::endl;
-		out << "AUTOSTART=" << stats.autostart << ::endl;
-		out << QString("IMPORTED=%1").arg(stats.imported_bytes) << ::endl;
+		st.write("PRIORITY", QString("%1").arg(priority));
+		st.write("AUTOSTART", QString("%1").arg(stats.autostart));
+		st.write("IMPORTED", QString("%1").arg(stats.imported_bytes));
+		st.write("CUSTOM_OUTPUT_NAME",custom_output_name ? "1" : "0");
+		st.write("MAX_RATIO", QString("%1").arg(maxShareRatio,0,'f',2));
+		st.write("RESTART_DISK_PREALLOCATION",prealloc ? "1" : "0");
+		
+		st.writeSync();
 	}
 
 	void TorrentControl::loadStats()
 	{
-		QFile fptr(datadir + "stats");
-		if (!fptr.open(IO_ReadOnly))
-			return;
-
-		QTextStream in(&fptr);
-		while (!in.atEnd())
+		StatsFile st(datadir + "stats");
+		
+		Uint64 val = st.readUint64("UPLOADED");
+		// stats.session_bytes_uploaded will be calculated based upon prev_bytes_ul
+		// seeing that this will change here, we need to save it 
+		session_bytes_uploaded = stats.session_bytes_uploaded; 
+		prev_bytes_ul = val;
+		up->setBytesUploaded(val);
+		
+		this->running_time_dl = st.readULong("RUNNING_TIME_DL");
+		this->running_time_ul = st.readULong("RUNNING_TIME_UL");
+		outputdir = st.readString("OUTPUTDIR").stripWhiteSpace();
+		if (st.hasKey("CUSTOM_OUTPUT_NAME") && st.readULong("CUSTOM_OUTPUT_NAME") == 1)
 		{
-			QString line = in.readLine();
-			if (line.startsWith("UPLOADED="))
-			{
-				bool ok = true;
-				Uint64 val = line.mid(9).toULongLong(&ok);
-				if (ok)
-					up->setBytesUploaded(val);
-				else
-					Out() << "Warning : can't get bytes uploaded out of line : "
-					<< line << endl;
-				prev_bytes_ul = val;
-			}
-			else if (line.startsWith("RUNNING_TIME_DL="))
-			{
-				bool ok = true;
-				unsigned long val  = line.mid(16).toULong(&ok);
-				if(ok)
-					this->running_time_dl = val;
-				else
-					Out() << "Warning : can't get running time out of line : "
-					<< line << endl;
-			}
-			else if (line.startsWith("RUNNING_TIME_UL="))
-			{
-				bool ok = true;
-				unsigned long val = line.mid(16).toULong(&ok);
-				if(ok)
-					this->running_time_ul = val;
-				else
-					Out() << "Warning : can't get running time out of line : "
-					<< line << endl;
-			}
-			else if (line.startsWith("OUTPUTDIR="))
-			{
-				outputdir = line.mid(10).stripWhiteSpace();
-			}
-			else if (line.startsWith("PRIORITY="))
-			{
-				bool ok = true;
-				int p = line.mid(9).toInt(&ok);
-				if(ok)
-				{
-					priority = p;
-					stats.user_controlled = p == 0 ? true : false;
-				}
-				else
-					Out() << "Warning : Can't get priority out of line : "
-							<< line << endl;
-			}
-			else if (line.startsWith("AUTOSTART="))
-			{
-				bool ok = true;
-				int p = line.mid(10).toInt(&ok);
-				if(ok)
-					stats.autostart = (bool) p;
-				else
-				{
-					Out() << "Warning : Can't get autostart bit out of line : "
-							<< line << endl;
-					stats.autostart = true;
-				}
-			}
-			else if (line.startsWith("IMPORTED="))
-			{
-				bool ok = true;
-				Uint64 p = line.mid(9).toULongLong(&ok);
-				if(ok)
-					stats.imported_bytes = p;
-				else
-				{
-					Out() << "Warning : Can't get imported_bytes out of line : "
-							<< line << endl;
-				}
-			}
+			custom_output_name = true;
 		}
+		
+		setPriority(st.readInt("PRIORITY"));
+		stats.user_controlled = priority == 0 ? true : false;
+		stats.autostart = st.readBoolean("AUTOSTART");
+		
+		stats.imported_bytes = st.readUint64("IMPORTED");
+		float rat = st.readFloat("MAX_RATIO");
+		maxShareRatio = rat;
+		if (st.hasKey("RESTART_DISK_PREALLOCATION"))
+			prealloc = st.readString("RESTART_DISK_PREALLOCATION") == "1";
+		
+		return;
 	}
 
 	void TorrentControl::loadOutputDir()
 	{
-		QFile fptr(datadir + "stats");
-		if (!fptr.open(IO_ReadOnly))
+		StatsFile st(datadir + "stats");
+		if (!st.hasKey("OUTPUTDIR"))
 			return;
-
-		QTextStream in(&fptr);
-		while (!in.atEnd())
+		
+		outputdir = st.readString("OUTPUTDIR").stripWhiteSpace();
+		if (st.hasKey("CUSTOM_OUTPUT_NAME") && st.readULong("CUSTOM_OUTPUT_NAME") == 1)
 		{
-			QString line = in.readLine();
-			if (line.startsWith("OUTPUTDIR="))
-			{
-				outputdir = line.mid(10).stripWhiteSpace();
-				if (outputdir.length() > 0 && !outputdir.endsWith(bt::DirSeparator()))
-					outputdir += bt::DirSeparator();
-				return;
-			}
+			custom_output_name = true;
 		}
 	}
 
@@ -771,14 +903,41 @@ namespace bt
 		stats.upload_rate = up && stats.running ? up->uploadRate() : 0;
 		stats.download_rate = down && stats.running ? down->downloadRate() : 0;
 		stats.bytes_left = cman ? cman->bytesLeft() : 0;
+		stats.bytes_left_to_download = cman ? cman->bytesLeft() - cman->bytesExcluded() : 0;
 		stats.bytes_uploaded = up ? up->bytesUploaded() : 0;
 		stats.bytes_downloaded = down ? down->bytesDownloaded() : 0;
 		stats.total_chunks = cman ? cman->getNumChunks() : 0;
 		stats.num_chunks_downloaded = cman ? cman->getNumChunks() - cman->chunksExcluded() - cman->chunksLeft() : 0;
 		stats.num_chunks_excluded = cman ? cman->chunksExcluded() : 0;
+		stats.chunk_size = tor ? tor->getChunkSize() : 0;
 		stats.total_bytes_to_download = (tor && cman) ?	tor->getFileLength() - cman->bytesExcluded() : 0;
-		stats.session_bytes_downloaded = stats.bytes_downloaded - prev_bytes_dl;
-		stats.session_bytes_uploaded = stats.bytes_uploaded - prev_bytes_ul;
+		
+		
+		
+		if (stats.bytes_downloaded >= prev_bytes_dl)
+			stats.session_bytes_downloaded = stats.bytes_downloaded - prev_bytes_dl;
+		else
+			stats.session_bytes_downloaded = 0;
+					
+		if (stats.bytes_uploaded >= prev_bytes_ul)
+			stats.session_bytes_uploaded = (stats.bytes_uploaded - prev_bytes_ul) + session_bytes_uploaded;
+		else
+			stats.session_bytes_uploaded = session_bytes_uploaded;
+		/*
+			Safety check, it is possible that stats.bytes_downloaded gets subtracted in Downloader.
+			Which can cause stats.bytes_downloaded to be smaller the trk_prev_bytes_dl.
+			This can screw up your download ratio.
+		*/
+		if (stats.bytes_downloaded >= trk_prev_bytes_dl)
+			stats.trk_bytes_downloaded = stats.bytes_downloaded - trk_prev_bytes_dl;
+		else
+			stats.trk_bytes_downloaded = 0;
+		
+		if (stats.bytes_uploaded >= trk_prev_bytes_ul)
+			stats.trk_bytes_uploaded = stats.bytes_uploaded - trk_prev_bytes_ul;
+		else
+			stats.trk_bytes_uploaded = 0;
+		
 		getSeederInfo(stats.seeders_total,stats.seeders_connected_to);
 		getLeecherInfo(stats.leechers_total,stats.leechers_connected_to);
 	}
@@ -900,10 +1059,182 @@ namespace bt
 	{
 		priority = p;
 		stats.user_controlled = p == 0 ? true : false;
+		if(p)
+			stats.status = kt::QUEUED;
+		else
+			updateStatusMsg();
+		
 		saveStats();
 	}
+	
+	void TorrentControl::setMaxShareRatio(float ratio)
+	{
+		if(ratio == 1.00f)
+		{
+			if(maxShareRatio != ratio)
+				maxShareRatio = ratio;
+		}
+		else
+			maxShareRatio = ratio;
+		
+		saveStats();
+		emit maxRatioChanged(this);
+	}
+	
+	bool TorrentControl::overMaxRatio()
+	{
+		if(stats.completed && stats.bytes_uploaded != 0 && stats.bytes_downloaded != 0 && maxShareRatio > 0)
+		{
+			if(kt::ShareRatio(stats) >= maxShareRatio)
+				return true;
+		}
+		
+		return false;
+	}
+
+	
+	QString TorrentControl::statusToString() const
+	{
+		switch (stats.status)
+		{
+			case kt::NOT_STARTED :
+				return i18n("Not started");
+			case kt::COMPLETE :
+				return i18n("Completed");
+			case kt::SEEDING :
+				return i18n("Seeding");
+			case kt::DOWNLOADING:
+				return i18n("Downloading");
+			case kt::STALLED:
+				return i18n("Stalled");
+			case kt::STOPPED:
+				return i18n("Stopped");
+			case kt::ERROR :
+				return i18n("Error: ") + getShortErrorMessage(); 
+			case kt::ALLOCATING_DISKSPACE:
+				return i18n("Allocating diskspace");
+			case kt::QUEUED:
+				return i18n("Queued");
+			case kt::CHECKING_DATA:
+				return i18n("Checking data");
+		}
+		return QString::null;
+	}
+
+	TrackersList* TorrentControl::getTrackersList()
+	{
+		AnnounceList* list = tor->getAnnounceList();
+		if(!list)
+			list = tor->createAnnounceList();
+		
+		return list;
+	}
+	
+	TrackersList* TorrentControl::createTrackersList()
+	{
+		return tor->createAnnounceList();
+	}
+
+	void TorrentControl::onPortPacket(const QString & ip,Uint16 port)
+	{
+		if (Globals::instance().getDHT().isRunning() && !stats.priv_torrent)
+			Globals::instance().getDHT().portRecieved(ip,port);
+	}
+	
+	void TorrentControl::doDataCheck(bt::DataCheckerListener* lst,bool auto_import)
+	{
+		if (stats.status == kt::ALLOCATING_DISKSPACE)
+			return;
+		
+		DataChecker* dc = 0;
+		stats.status = kt::CHECKING_DATA;
+		if (stats.multi_file_torrent)
+			dc = new MultiDataChecker();
+		else
+			dc = new SingleDataChecker();
+	
+		dc->setListener(lst);
+		dc->check(stats.output_path,*tor,datadir + "dnd" + bt::DirSeparator());
+		
+		if (lst && !lst->isStopped())
+		{
+			down->dataChecked(dc->getDownloaded());
+			// update chunk manager
+			cman->dataChecked(dc->getDownloaded());
+			if (auto_import)
+			{
+				down->recalcDownloaded();
+				stats.imported_bytes = down->bytesDownloaded();
+				if (cman->haveAllChunks())
+					stats.completed = true;
+			}
+		}
+		delete dc;
+		// update the status
+		updateStatusMsg();
+		updateStats();
+	}
+	
+	bool TorrentControl::hasExistingFiles() const
+	{
+		return cman->hasExistingFiles();
+	}
+	
+	bool TorrentControl::hasMissingFiles(QStringList & sl)
+	{
+		return cman->hasMissingFiles(sl);
+	}
+	
+	void TorrentControl::recreateMissingFiles()
+	{
+		try
+		{
+			cman->recreateMissingFiles();
+			prealloc = true; // set prealloc to true so files will be truncated again
+			down->dataChecked(cman->getBitSet()); // update chunk selector
+		}
+		catch (Error & err)
+		{
+			onIOError(err.toString());
+			throw;
+		}
+	}
+	
+	void TorrentControl::dndMissingFiles()
+	{
+		try
+		{
+			cman->dndMissingFiles();
+			prealloc = true; // set prealloc to true so files will be truncated again
+			missingFilesMarkedDND(this);
+			down->dataChecked(cman->getBitSet()); // update chunk selector
+		}
+		catch (Error & err)
+		{
+			onIOError(err.toString());
+			throw;
+		}
+	}
+	
+	void TorrentControl::handleError(const QString & err)
+	{
+		onIOError(err);
+	}
+	
+	Uint32 TorrentControl::getNumDHTNodes() const
+	{
+		return tor->getNumDHTNodes();
+	}
+	
+	const kt::DHTNode & TorrentControl::getDHTNode(Uint32 i) const 
+	{
+		return tor->getDHTNode(i);
+	}
+
+	void TorrentControl::deleteDataFiles()
+	{
+		cman->deleteDataFiles();
+	}
 }
 
-
-		
 #include "torrentcontrol.moc"
