@@ -15,11 +15,12 @@
  *   You should have received a copy of the GNU General Public License     *
  *   along with this program; if not, write to the                         *
  *   Free Software Foundation, Inc.,                                       *
- *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
+ *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.             *
  ***************************************************************************/
 #include <klocale.h>
 #include <kglobal.h>
 #include <kiconloader.h>
+#include <kstandarddirs.h>
 #include <kmessagebox.h>
 #include <ksocketaddress.h>
 #include <qpoint.h>
@@ -29,47 +30,134 @@
 #include <interfaces/functions.h>
 #include <torrent/ipblocklist.h>
 #include "peerview.h"
-
+#include "GeoIP.h"
 
 using namespace bt;
 using namespace kt;
 
 namespace kt
 {
+	
+	Uint32 PeerViewItem::pvi_count = 0;
+	// Global GeoIP pointer, gets destroyed when no PeerViewItem's exist
+	static GeoIP* geo_ip = 0; 
+	static QPixmap yes_pix;
+	static QPixmap no_pix;
+	static QPixmap lock_pix;
+	static bool yes_no_pix_loaded = false;
+	static bool geoip_db_exists = true;
+	
 		
 	PeerViewItem::PeerViewItem(PeerView* pv,kt::PeerInterface* peer) : KListViewItem(pv),peer(peer)
-	{
+	{	
+		if (!yes_no_pix_loaded)
+		{
+			KIconLoader* iload = KGlobal::iconLoader();
+			yes_pix = iload->loadIcon("button_ok",KIcon::Small);
+			no_pix = iload->loadIcon("button_cancel",KIcon::Small);
+			lock_pix = iload->loadIcon("ktencrypted",KIcon::Small);
+			geoip_db_exists = !locate("data", "ktorrent/geoip/geoip.dat").isNull();
+			yes_no_pix_loaded = true;
+		}
+		
+		pvi_count++;
+		const char * hostname = 0;
+		const char * country_code = 0;
+		const char * country_name = 0;
+		int country_id = 0;
+
+		const PeerInterface::Stats & s = peer->getStats();
+		hostname = s.ip_addresss.ascii();
+
+		// open GeoIP if necessary
+		if (!geo_ip && geoip_db_exists)
+			geo_ip = GeoIP_open(locate("data", "ktorrent/geoip/geoip.dat").ascii(),0);
+		
+		if (geo_ip)
+		{
+			country_id = GeoIP_id_by_name(geo_ip, hostname);
+			country_code = GeoIP_country_code[country_id];
+			country_name = GeoIP_country_name[country_id];
+			setText(1, country_name);
+			m_country = QString(country_name);
+		}
+		else
+		{
+			setText(1,"N/A");
+		}
+		
+		setText(0,s.ip_addresss);
+		setText(2,s.client);
+		
+		if (country_code)
+		{
+			QPixmap pix(locate("data", 	QString("ktorrent/geoip/%1.png").arg(QString(country_code)).lower()));
+			setPixmap(1, pix);
+		}
+		
+		if (s.encrypted)
+			setPixmap(0,lock_pix);
 		update();
 	}
 	
+	PeerViewItem::~PeerViewItem()
+	{
+		if (pvi_count > 0) // just to be sure, let this not wrap around
+			pvi_count--;
+		
+		// destroy when not needed anymore
+		if (pvi_count == 0 && geo_ip)
+		{
+			GeoIP_delete(geo_ip);
+			geo_ip = 0;
+		}
+	}
+	
 	
 	void PeerViewItem::update()
 	{
 		KLocale* loc = KGlobal::locale();
 		const PeerInterface::Stats & s = peer->getStats();
-		setText(0,s.ip_addresss);
-		setText(1,s.client);
-		setText(2,KBytesPerSecToString(s.download_rate / 1024.0));
-		setText(3,KBytesPerSecToString(s.upload_rate / 1024.0));
-		setText(4,s.choked ? i18n("yes") : i18n("no"));
-		setText(5,s.snubbed ? i18n("yes") : i18n("no"));
-		setText(6,QString("%1 %").arg(loc->formatNumber(s.perc_of_file,2)));
+		
+		setText(3,KBytesPerSecToString(s.download_rate / 1024.0));
+		setText(4,KBytesPerSecToString(s.upload_rate / 1024.0));
+		//setPixmap(5,!s.choked ? yes_pix : no_pix);
+		setText(5,s.choked ? i18n("Yes") : i18n("No"));
+		//setPixmap(6,!s.snubbed ? yes_pix : no_pix);
+		setText(6,s.snubbed ? i18n("Yes") : i18n("No"));
+		setText(7,QString("%1 %").arg(loc->formatNumber(s.perc_of_file,2)));
+		setPixmap(8,s.dht_support ? yes_pix : no_pix);
+		setText(9,loc->formatNumber(s.aca_score,2));
+		setPixmap(10,s.has_upload_slot ? yes_pix : QPixmap());
+#undef SHOW_REQUESTS
+#ifdef SHOW_REQUESTS
+		setText(11,QString("%1 / %2").arg(s.num_down_requests).arg(s.num_up_requests));
+#endif
 	}
 	
 	int PeerViewItem::compare(QListViewItem * i,int col,bool) const
 	{
-		PeerInterface* op = ((PeerViewItem*)i)->peer;
+		PeerViewItem* pvi = (PeerViewItem*) i;
+		PeerInterface* op = pvi->peer;
 		const PeerInterface::Stats & s = peer->getStats();
 		const PeerInterface::Stats & os = op->getStats();
 		switch (col)
 		{
 			case 0: return QString::compare(s.ip_addresss,os.ip_addresss);
-			case 1: return QString::compare(s.client,os.client);
-			case 2: return CompareVal(s.download_rate,os.download_rate);
-			case 3: return CompareVal(s.upload_rate,os.upload_rate);
-			case 4: return CompareVal(s.choked,os.choked);
-			case 5: return CompareVal(s.snubbed,os.snubbed);
-			case 6: return CompareVal(s.perc_of_file,os.perc_of_file);
+			case 1:	
+			{
+				int ret = QString::compare(m_country, pvi->m_country);
+				return ret;
+			}	
+			case 2: return QString::compare(s.client,os.client);
+			case 3: return CompareVal(s.download_rate,os.download_rate);
+			case 4: return CompareVal(s.upload_rate,os.upload_rate);
+			case 5: return CompareVal(s.choked,os.choked);
+			case 6: return CompareVal(s.snubbed,os.snubbed);
+			case 7: return CompareVal(s.perc_of_file,os.perc_of_file);
+			case 8: return CompareVal(s.dht_support,os.dht_support);
+			case 9: return CompareVal(s.aca_score,os.aca_score);
+			case 10: return CompareVal(s.has_upload_slot,os.has_upload_slot);
 		}
 		return 0;
 	}
@@ -78,12 +166,19 @@ namespace kt
 			: KListView(parent, name)
 	{
 		addColumn(i18n("IP"));
+		addColumn(i18n("Country"));
 		addColumn(i18n("Client"));
 		addColumn(i18n("Down Speed"));
 		addColumn(i18n("Up Speed"));
 		addColumn(i18n("Choked"));
 		addColumn(i18n("Snubbed"));
 		addColumn(i18n("Availability"));
+		addColumn(i18n("DHT"));
+		addColumn(i18n("Score"));
+		addColumn(i18n("Upload Slot"));
+#ifdef SHOW_REQUESTS
+		addColumn(i18n("Requests"));
+#endif
 		setShowSortIndicator(true);
 		
 		menu = new KPopupMenu(this);
@@ -106,9 +201,19 @@ namespace kt
 	
 	void PeerView::removePeer(kt::PeerInterface* peer)
 	{
-		PeerViewItem* it = items[peer];
-		delete it;
+		QMap<kt::PeerInterface*,PeerViewItem*>::iterator it = items.find(peer);
+		if (it == items.end())
+		{
+			return;
+		}
+
+		PeerViewItem* pvi = it.data();
+		if (pvi == curr)
+			curr = 0;
+				  
+		delete pvi;
 		items.erase(peer);
+		
 	}
 	
 	void PeerView::banPeer(kt::PeerInterface* peer)
